/*
 * redir.c - Provide a transparent TCP proxy through remote shadowsocks
 *           server
 *
 * Copyright (C) 2013 - 2016, Max Lv <max.c.lv@gmail.com>
 *
 * This file is part of the shadowsocks-libev.
 *
 * shadowsocks-libev is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * shadowsocks-libev is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with shadowsocks-libev; see the file COPYING. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <limits.h>
#include <linux/if.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6/ip6_tables.h>
#include <udns.h>
#include <stdio.h>

#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>

//#include <netinet/tcp.h>
#include <sys/un.h>
//#include <sys/socket.h>
//#include <pwd.h>


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ev.h>

#include "pear_base64.h"
#include "pear_utils.h"
#include "pear_json.h"
#include "pear_jconf.h"

#include "../utils.h"
#include "pear_log.h"
#include "pear_host.h"
#include "pear_curl.h"
#include "pear_fog_info.h"
#include "sodium/crypto_hash_sha256.h"
#include "sodium/crypto_auth_hmacsha256.h"

#ifndef EAGAIN
#define EAGAIN EWOULDBLOCK
#endif

#ifndef EWOULDBLOCK
#define EWOULDBLOCK EAGAIN
#endif

#ifndef BUF_SIZE
#define BUF_SIZE 2048
#endif

#ifndef IP6T_SO_ORIGINAL_DST
#define IP6T_SO_ORIGINAL_DST 80
#endif


#define PEAR_LIST_UPDATE_TIME    (60 * 60 * 24 * 30)
#define PEAR_ROUTER_REPORT_TIME  (60 * 5)

#define PEAR_CHANNEL_CHECK_TIME  (60 * 2)
#define PEAR_CHANNEL_CHECK_SLEEP 1
#define PEAR_CHANNEL_CHECK_NUM   3

#define PEAR_CHANNEL_CHECK_URL "https://www.google.com"
static void pear_channel_restart(EV_P_ ev_timer *w, int revents);


#define PEAR_ARRAY_SIZE(array) (sizeof(array)/sizeof(array[0]))

struct pear_ev_s {
    struct ev_loop *pear_loop;
    struct ev_timer pear_timeout_channel_check;
    struct ev_timer pear_timeout_channel_restart;
};
struct pear_ev_s pear_ev;


char *shadowfog_router_host_address[] = {
    "https://downloads.shadowfog.com/host/sf-host.json",
    "https://git.oschina.net/ShadowFogNetwork/Information/raw/master/ss-manager.json",
    "https://host.ablive.net/host.json"
};


struct pear_config {
    int   mtu;
    int   mptcp;
    int   nofile;
    int   pid_flags;
    int   pear_mode;
    int   remote_num;

    char *user;
    char *server;
    char *method;
    char *timeout;
    char *pid_path;
    char *password;
    char *pear_user;
    char *conf_path;
    char *list_ipset;
    char *local_addr;
    char *remote_dns;
    char *redir_port;
    char *tunnel_port;
    char *pear_passwd;
    char *remote_port;
    char *remote_host;
    char *pear_dnsurl;
    char *domain_file;
    char *pear_dnsmasq;
};
struct pear_config config;

struct pear_channel {
    char port[16];
    char server[32];
    char passwd[128];
};
struct pear_channel cloud_channel;

struct pear_client {
    int  type;
    char pidfile[128];
};

struct pear_client client_redir;
struct pear_client client_tunnel;

typedef struct {
    int  fd;
    int  stat;
    char message[152];
} pear_stat_t;

pear_stat_t pear_stat;

//char tempfile[] = "/tmp/ShadowFogStat";
char tempfile[] = "/etc/config/ShadowFogStat";


static char *pear_curl_header[] = {
    "Content-Type: application/json",
    "Cache-Control: no-chche",
     NULL
};

char working_dir[] = "/var/run";
#define CHANNEL_HOST_NUM  2

char *shadowfog_channel_host[] = {
    "https://downloads.shadowfog.com/host/sf-scheduler.json",
    "http://git.oschina.net/ShadowFogNetwork/Information/raw/master/sf-scheduler.json",
    "https://host.ablive.net/scheduler.json"
};

void pear_stop_client(struct pear_client *);
int  pear_get_channel(char *, char *, struct pear_channel *);
void pear_start_client(struct pear_client *, struct pear_channel *);
static int  pear_channel_check_connect(struct pear_channel *channel);

void pear_stat_exit()
{
    LOGI("[ss-router] ==========================pear_stat_exit =======================");

    if (ftruncate(pear_stat.fd, 0) < 0)
    {
        LOGI("[ss-router] ftruncate error!");
    }

    unlink(tempfile);
}

void pear_stat_init()
{
    unlink(tempfile);

    if (atexit(pear_stat_exit) < 0)
    {
        LOGI("[ss-router] atexit error: %s", strerror(errno));
    }

    pear_stat.fd = open(tempfile, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
    if (pear_stat.fd < 0)
    {
        LOGI("[ss-router] open error: %s", strerror(errno));
    }

}

void pear_stat_output()
{
    char info[256] = {0x00};

    LOGI("[ss-router] pear_stat_output! pear_stat.stat: %d, pear_stat.message:%s", pear_stat.stat, pear_stat.message);

/*
    sprintf(info, "{"
                    "\"stat\":\"%d\","
                    "\"message\":\"%s\""
                  "}",
                  pear_stat.stat,
                  pear_stat.message);
*/

    sprintf(info, "config stat message\n"
                    "option stat \"%d\"\n"
                    "option message \"%s\"\n",
                  pear_stat.stat,
                  pear_stat.message);

    if (ftruncate(pear_stat.fd, 0) < 0)
    {
        LOGI("[ss-router] ftruncate error!");
    }

    if (lseek(pear_stat.fd, 0, SEEK_SET) < 0)
    {
        LOGI("[ss-router] lseek error!");
    }
    if (write(pear_stat.fd, info, strlen(info)) < 0)
    {
        LOGI("[ss-router] write error!");
    }
}

void pear_stat_list()
{
    pear_stat.stat = -1;
    strcpy(pear_stat.message, "DNSList SYNC Failed. Please check your network!");
    pear_stat_output();
}

void pear_stat_check(int result)
{
    if (result == 1)
    {
        pear_stat.stat = 1;
        memset(pear_stat.message, 0x00, sizeof(pear_stat.message));
    }
    else
    {
        pear_stat.stat = -1;
    }

    pear_stat_output();
}

void pear_stat_connecting()
{
    pear_stat.stat = 0;
    strcpy(pear_stat.message, "Connecting to Fog...");
    pear_stat_output();
}

void pear_stat_host()
{
    pear_stat.stat = -1;
    strcpy(pear_stat.message, "The network is unreachable!");
    pear_stat_output();
}

void pear_stat_channel(int result, char *message)
{
    char temp[256] = {0x00};

    if (result == 1)
    {
        pear_stat.stat = 0;
        memset(pear_stat.message, 0x00, sizeof(pear_stat.message));
    }
    else
    {
        pear_stat.stat = -1;
        pear_json_get_string(message, "errormsg" ,temp);
        strcpy(pear_stat.message, temp);
    }

    pear_stat_output();
}

void signal_cb(int dummy)
{
    LOGI("[ss-router] signal catched: %d!", dummy);
    exit(-1);
}

void pear_set_ipset(struct pear_channel *channel)
{
/*
    ipset_init_libarary();
    struct ip_set server_ip;
    ipset_init(&server_ip);
    ipset_ipv4_add(&server_ip, &ip);
*/

    char cmd[128] = {0x00};
    sprintf(cmd, "ipset create server_set hash:ip; ipset add server_set %s;", channel->server);
    LOGI("[pear-router]pear_set_ipset cmd: %s", cmd);
    if (system(cmd) == -1)
    {
        LOGI("[pear-router] pear_set_ipset error");
    }
}

static int pear_channel_check_connect(struct pear_channel *channel)
{
    LOGI("[router] ================ check connect start ===================\n");
    int n     = 0;
    int fd    = 0;
    int error = 0;
    int flags = 0;

    socklen_t len;
    fd_set rset, wset;

    struct timeval     tval;
    struct sockaddr_in addr;

    addr.sin_family = AF_INET;
    addr.sin_port   = htons(atoi(channel->port));
    inet_pton(AF_INET, channel->server, &addr.sin_addr);
    LOGI("port :%s \t server: %s\n", channel->port, channel->server);
    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0)
    {
        LOGI("Function socket() Error!\n");
        return -1;
    }

    flags = fcntl(fd, F_GETFL, 0);
    fcntl(fd, F_SETFL, flags | O_NONBLOCK);

    if ( (n = connect(fd, (struct sockaddr *)&addr, sizeof(addr))) < 0)
    {
        if (errno != EINPROGRESS)
        {
            LOGI("[router] Function connect ERROR!\n");
            return -1;
        }
    }

    if (n == 0)
    {
        LOGI("[channel-check] connected");
        close(fd);
        return 0;
    }

    FD_ZERO(&rset);
    FD_SET(fd, &rset);
    wset = rset;
    tval.tv_sec  = 3;
    tval.tv_usec = 0;

    if ( (n = select(fd + 1, &rset, &wset, NULL, &tval)) <= 0)
    {
        close(fd);
        LOGI("[router] connected timeout");
        return -1;
    }

    error = 0;
    if (FD_ISSET(fd, &rset) || FD_ISSET(fd, &wset))
    {
        len = sizeof(error);
        if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) < 0)
        {
            LOGI("[channel-check] getsockopt error: %d\n", error);
            close(fd);
            return -1;
        }

        if (error)
        {
            LOGI("[channel-check] error: %d\n", error);
            close(fd);
            return -1;
        }

        LOGI("[router] connected OK!");
        close(fd);
        return 0;
    }

    LOGI("[channel-check] Connected After Select NOT IN SET");
    return -1;
}

int pear_channel_check(struct pear_channel *channel)
{
    int   s = 0;
    char recvbuf[2048] = {0x00};

    s = pear_curl_cmd("GET", PEAR_CHANNEL_CHECK_URL, NULL, pear_curl_header, NULL, 0, recvbuf);
    LOGI("[pear-router] pear_curl_cmd: %s, readbyte: %d", PEAR_CHANNEL_CHECK_URL, s);

    if (s >= 0)
    {
        pear_stat_check(1);
        return 0;
    }
    pear_stat_check(-1);

    return 1;
}

/*
static void pear_get_list()
{
    int  i = 0;
    char cmd[256] = {0x00};

    sprintf(cmd, "ss-list2dns -o %s --list-url %s -s %s -d %s -p %s -i --extra-domain-file %s",
            config.pear_dnsmasq,
            config.pear_dnsurl,
            config.list_ipset,
            "127.0.0.1",
            config.tunnel_port,
            (config.domain_file == NULL) ? "/tmp/domain" : config.domain_file);

    LOGI("[ss-router] system cmd: %s\n", cmd);

    for (i = 1; ; i++, i = i % 100)
    {

        if (system(cmd) != 0)
        {
            LOGI("[ss-router] Update List failed!");
            pear_stat_list();
            LOGI("[ss-router] Get list after sleep:%d", i);
            sleep(i);
            continue;
        }
 
        if (system("/etc/init.d/dnsmasq restart") == 0)
        {
            LOGI("[ss-router] Update List success!");
            break;
        }
        else
        {
            LOGI("[ss-router] Update List failed!");
        }
    }
}
*/

static void pear_timeout_report_cb(EV_P_ ev_timer *w, int revents)
{
    LOGI("[sf-router] ===========pear_timeout_report_cb starting...=============");
    int  s = 0;
    char report_buf[1024]     = {0x00};
    char host_address[1024]   = {0x00};
    char response_buf[2048]   = {0x00};
    char report_address[1024] = {0x00};

    pear_fog_info_detect();

    pear_get_host(host_address, shadowfog_router_host_address,
                  PEAR_ARRAY_SIZE(shadowfog_router_host_address), "shadowfog_report");


    sprintf(report_address, "%s/report", host_address);
    LOGI("[sf-router] host_address:%s", report_address);

    sprintf(report_buf, "{"
                            " \"mac\":             \"%s\", "
                            " \"is_public_ip\":    \"%d\", "
                            " \"is_support_pmp\":  \"%d\", "
                            " \"is_support_upnp\": \"%d\", "
                            " \"local_ip\":        \"%s\", "
                            " \"extern_ip\":       \"%s\", "
                            " \"cpu_number\":      \"%d\", "
                            " \"psn\":             \"%s\", "
                            " \"memory\":          \"%s\", "
                            " \"net_type\":        \"%d\", "
                            " \"platform\":        \"%s\", "
                            " \"shell_result\":    \"%s\"  "
                        "}",
                            pear_fog_info.mac,
                            pear_fog_info.is_public_ip,
                            pear_fog_info.is_support_pmp,
                            pear_fog_info.is_support_upnp,
                            pear_fog_info.local_ip,
                            pear_fog_info.extern_ip,
                            pear_fog_info.cpu_number,
                            pear_fog_info.psn,
                            pear_fog_info.memery,
                            pear_fog_info.net_type,
                            pear_fog_info.platform,
                            pear_fog_info.shellresult);

    LOGI("[sf-router] report_buf:%s", report_buf);

    s = pear_curl_cmd("POST", report_address, NULL, pear_curl_header, report_buf, strlen(report_buf), response_buf);
    if (s < 0)
    {
        LOGI("[sf-router] pear_curl_cmd error");
    }
    else
    {
        // Base64decode(base64_temp, response_buf);
        LOGI("[sf-router] response_buf: %s", response_buf);
    }

    LOGI("[sf-router] ===========pear_timeout_report_cb END...=============");
}

static void pear_timeout_dnslist_cb(EV_P_ ev_timer *w, int revents)
{
    char cmd[256] = {0x00};
    sprintf(cmd, "ss-list2dns -o %s --list-url %s -s %s -d %s -p %s -i --extra-domain-file %s", config.pear_dnsmasq, config.pear_dnsurl, config.list_ipset, "127.0.0.1", config.tunnel_port, (config.domain_file == NULL) ? "/tmp/domain" : config.domain_file);

    LOGI("[ss-router] system cmd: %s\n", cmd);

    if (system(cmd) != 0)
    {
        LOGI("[ss-router] Update List failed!");
        return;
    }

    if (system("/etc/init.d/dnsmasq restart") != 0)
    {
        LOGI("[ss-router] DNS Restart failed!");
    }
}

static void pear_channel_check_cb(EV_P_ ev_timer *w, int revents)
{
    LOGI("[pear-router]pear channel check start \n");
    int i = 0;

    for (i = 0; i < PEAR_CHANNEL_CHECK_NUM; i++)
    {
        if (pear_channel_check_connect(&cloud_channel) == 0)
        {
            LOGI("[pear-router] pear channel success!");
            return;
        }
        LOGI("[pear-router] channel check again: %d", i);
        //sleep(PEAR_CHANNEL_CHECK_SLEEP);
    }
    LOGI("[pear-router]pear channel invalid");

    ev_timer_stop(pear_ev.pear_loop, &(pear_ev.pear_timeout_channel_check));
    ev_timer_init(&(pear_ev.pear_timeout_channel_restart), pear_channel_restart, 0, 0);
    ev_timer_start(pear_ev.pear_loop, &(pear_ev.pear_timeout_channel_restart));
}

static void pear_channel_restart(EV_P_ ev_timer *w, int revents)
{
    LOGI("[sf-router] ============ pear_channel_restart Starting... ==============");

    if (pear_get_channel(config.pear_user, config.pear_passwd, &cloud_channel) < 0)
    {
        ev_timer_init(&(pear_ev.pear_timeout_channel_restart), pear_channel_restart, 5, 0);
        ev_timer_start(pear_ev.pear_loop, &(pear_ev.pear_timeout_channel_restart));

        return;
    }

    pear_set_ipset(&cloud_channel);

    pear_stop_client(&client_redir);
    pear_stop_client(&client_tunnel);

    pear_start_client(&client_redir,  &cloud_channel);
    pear_start_client(&client_tunnel, &cloud_channel);

//    ev_timer_stop(pear_loop, &pear_timeout_channel_restart);
    ev_timer_start(pear_ev.pear_loop, &(pear_ev.pear_timeout_channel_check));
    LOGI("[sf-router] ============ pear_channel_restart END... ==============");
}

void pear_start_client(struct pear_client *client, struct pear_channel *channel)
{
    char cmd[256] = {0x00};

    if (client->type == 0)
    {
        sprintf(cmd, "%s -s %s -p %s -k %s -b \"0.0.0.0\" -l %s -m %s -f %s", "ss-redir", channel->server, channel->port, channel->passwd, config.redir_port, config.method, client->pidfile);
    }
    else if (client->type == 1)
    {
        sprintf(cmd, "%s -s %s -p %s -k %s -b \"0.0.0.0\" -l %s -m %s -f %s -L %s -u", "ss-tunnel", channel->server, channel->port, channel->passwd, config.tunnel_port, config.method, client->pidfile, config.remote_dns);
    }
    else
    {
        LOGE("[pear-router] unknow client type: %d", client->type);
        return;
    }

    LOGI("[pear-router]system cmd: %s", cmd);
    if (system(cmd) == -1)
    {
        LOGI("[pear-router] pear_start_client system error type: %d", client->type);
    }

}

void pear_stop_client(struct pear_client *client)
{
    int pid;

    FILE *f = fopen(client->pidfile, "r");
    if (f == NULL)
    {
        LOGE("[pear-r-t]unable to open pidfile");
        return;
    }

    if (fscanf(f, "%d", &pid) != EOF)
    {
        kill(pid, SIGTERM);
    }

    fclose(f);
}

static void pear_BIN_to_XX(unsigned char *bin, char *xx_str, int bin_len)
{
    int i;

    for (i = 0; i < bin_len; i++)
    {
        sprintf((char *)xx_str + i*2, "%02x", (unsigned char)bin[i]);
    }
}

int pear_hmac_sha256(char *signature, int signature_len, char *pear_user, char *pear_passwd, char *timestamp, char *nonce)
{
    char key[512]      = {0x00};
    char value[256]    = {0x00};
    char key_str[128]  = {0x00};
    char sign_bin[512] = {0x00};

    crypto_hash_sha256((unsigned char *)key, (unsigned char *)pear_passwd, strlen(pear_passwd));
    pear_BIN_to_XX((unsigned char *)key, key_str, 32);
    LOGI("[pear-router]key_str: %d, %s", crypto_auth_hmacsha256_KEYBYTES, key_str);

    sprintf(value, "shadowfog%s%s%s", timestamp, nonce, pear_user);
    LOGI("[pear-router]value: %s", value);

    crypto_auth_hmacsha256_state state;
    crypto_auth_hmacsha256_init(&state, (unsigned char *)key_str, strlen(key_str));

    crypto_auth_hmacsha256_update(&state, (unsigned char *)value, strlen(value));

    crypto_auth_hmacsha256_final(&state, (unsigned char *)sign_bin);
    pear_BIN_to_XX((unsigned char *)sign_bin, signature, 32);
    LOGI("[pear-router]signature: %s", signature);

    return 0;
}

int pear_get_channel(char *pear_user, char *pear_passwd, struct pear_channel *channel)
{
    int  i = 0;
    int  s = 0;
    char cmd[1024]           = {0x00};
    char temp[4096]          = {0x00};

    char timestamp[20]       = {0x00};
    char signature[128]      = {0x00};
    char scheduler[1024]     = {0x00};
    char base64_decode[1024] = {0x00};

    char nonce[]             = "leslie";
    char transactionID[]     = "backstring";

    for (i = 0; i < CHANNEL_HOST_NUM; i++)
    {
        sleep(1);

        LOGI("[pear-router] pear_curl_cmd: %s", shadowfog_channel_host[i]);

        s = pear_curl_cmd("GET", shadowfog_channel_host[i], NULL, pear_curl_header, NULL, 0, temp);
        if (s < 0)
        {
            pear_stat_host();
            continue;
        }

        LOGI("[pear-router] Read feedback bytes %d , and content : %s", s, temp);

        Base64decode(base64_decode, temp);

        LOGI("[pear-router] After base64 decode the  content : %s", base64_decode);

        if (strstr(base64_decode, "scheduler_addr") == NULL)
        {
            pear_stat_host();
            continue;
        }

        pear_json_get_string(base64_decode, "scheduler_addr", scheduler);
        LOGI("[pear-router]scheduler:%s", scheduler);

        memset(temp, 0x00, 4096);
        sprintf(timestamp, "%ld", time(NULL));
        LOGI("[pear-router]timestamp: %s", timestamp);

        pear_hmac_sha256(signature, sizeof(signature), pear_user, pear_passwd, timestamp, nonce);

        sprintf(cmd, "%s?username=%s&timestamp=%s&nonce=%s&transactionID=%s&signature=%s", scheduler, pear_user, timestamp, nonce, transactionID, signature);
        s = pear_curl_cmd("GET", cmd, NULL, pear_curl_header, NULL, 0, temp);
        if (s < 0)
        {
            pear_stat_host();
            continue;
        }

        LOGI("[pear-router] pear_curl_cmd:%s\n, read feedback bytes %d, and content : %s", cmd, s, temp);

        if (strstr(temp, "server_port") == NULL)
        {
            pear_stat_channel(0, temp);
            LOGI("[pear-router] Get channel error sleep 5S and require again");
            sleep(5);
            continue;
        }

        pear_stat_channel(1, NULL);
        pear_json_get_server(temp, channel->server, channel->port, channel->passwd);
        LOGI("[pear-router] server: %s, port: %s, passwd:%s", channel->server, channel->port, channel->passwd);

        return 0;
    }

    return -1;
}

int main(int argc, char **argv)
{
    int   c;

    pear_stat_init();
    pear_stat_connecting();

    srand(time(NULL));
    memset(&config, 0x00, sizeof(config));
/*
    int   mtu          = 0;
    int   mptcp        = 0;
    int   pid_flags    = 0;
    int   pear_mode    = 0;
    int   remote_num   = 0;

    char *user         = NULL;
    char *method       = NULL;
    char *timeout      = NULL;
    char *pid_path     = NULL;
    char *password     = NULL;
    char *pear_user    = NULL;
    char *conf_path    = NULL;
    char *local_port   = NULL;
    char *local_addr   = NULL;
    char *remote_dns   = NULL;
    char *pear_passwd  = NULL;
    char *remote_port  = NULL;

    char port[10]      = {0x00};
    char server[32]    = {0x00};
    char passwd[128]   = {0x00};
*/

    int option_index                    = 0;
    static struct option long_options[] = {
        { "mtu",          required_argument, 0, 0 },
        { "mptcp",        no_argument,       0, 0 },
        { "pear_mode",    no_argument,       0, 0 },
        { "pear_user",    required_argument, 0, 0 },
        { "redir_port",   required_argument, 0, 0 },
        { "tunnel_port",  required_argument, 0, 0 },
        { "pear_passwd",  required_argument, 0, 0 },
        { "pear_dnsmasq", required_argument, 0, 0 },
        { "pear_dnsurl",  required_argument, 0, 0 },
        { "list_ipset",   required_argument, 0, 0 },
        { "domain_file",  required_argument, 0, 0 },
        { "help",         no_argument,       0, 0 },
        {      0,         0,                 0, 0 }
    };

    opterr = 0;

    USE_TTY();

    while ((c = getopt_long(argc, argv, "f:s:p:L:k:t:m:c:b:a:n:huUvA",
                            long_options, &option_index)) != -1) {
        switch (c) {
        case 0:
            if (option_index == 0) {
                config.mtu = atoi(optarg);
                LOGI("set MTU to %d", config.mtu);
            } else if (option_index == 1) {
                config.mptcp = 1;
                LOGI("enable multipath TCP");
            } else if (option_index == 2) {
                config.pear_mode = 1;
            } else if (option_index == 3) {
                config.pear_user = optarg;
            } else if (option_index == 4) {
                config.redir_port = optarg;
            } else if (option_index == 5) {
                config.tunnel_port = optarg;
            } else if (option_index == 6) {
                config.pear_passwd = optarg;
            } else if (option_index == 7) {
                config.pear_dnsmasq = optarg;
            } else if (option_index == 8) {
                config.pear_dnsurl = optarg;
            } else if (option_index == 9) {
                config.list_ipset = optarg;
            } else if (option_index == 10) {
                config.domain_file = optarg;
            } else if (option_index == 11) {
                usage();
                exit(EXIT_SUCCESS);
            }
            break;
        case 's':
                config.remote_host   = optarg;
            break;
        case 'p':
            config.remote_port = optarg;
            break;
        case 'L':
            config.remote_dns = optarg;
            break;
        case 'k':
            config.password   = optarg;
            break;
        case 'f':
            config.pid_flags  = 1;
            config.pid_path   = optarg;
            break;
        case 't':
            config.timeout    = optarg;
            break;
        case 'm':
            config.method     = optarg;
            break;
        case 'c':
            config.conf_path  = optarg;
            break;
        case 'b':
            config.local_addr = optarg;
            break;
        case 'a':
            config.user = optarg;
            break;
#ifdef HAVE_SETRLIMIT
        case 'n':
            config.nofile = atoi(optarg);
            break;

#endif
/*
        case 'u':
            config.mode = TCP_AND_UDP;
            break;
        case 'U':
            config.mode = UDP_ONLY;
            break;
        case 'v':
            config.verbose = 1;
            break;
        case 'A':
            config.auth = 1;
            break;
*/
        case 'h':
            usage();
            exit(EXIT_SUCCESS);
        case '?':
            // The option character is not recognized.
            LOGE("Unrecognized option: %s", optarg);
            opterr = 1;
            break;
        }
    }

    router_config_t *json_config = NULL;

    if (config.conf_path != NULL)
    {
        json_config = pear_json_read_conf(config.conf_path);
    }

    if (json_config != NULL)
    {
        LOGI("[router] cp json_config start");
        config.mtu          = (json_config->mtu    == NULL) ? 0 : atoi(json_config->mtu);
        config.mptcp        = (json_config->mptcp  == NULL) ? 0 : atoi(json_config->mptcp);
        config.nofile       = (json_config->nofile == NULL) ? 0 : atoi(json_config->nofile);
        config.pear_mode    = (json_config->pear_mode == NULL) ? 0 : atoi(json_config->pear_mode);

        config.user         = json_config->user;
        config.method       = json_config->method;
        config.timeout      = json_config->timeout;
        config.pid_path     = json_config->pid_path;
        config.password     = json_config->password;

        config.pear_user    = json_config->pear_user;
        config.list_ipset   = json_config->list_ipset;
        config.local_addr   = json_config->local_addr;
        config.remote_dns   = json_config->remote_dns;
        config.redir_port   = json_config->redir_port;
        config.tunnel_port  = json_config->tunnel_port;
        config.pear_passwd  = json_config->pear_passwd;
        config.remote_port  = json_config->remote_port;
        config.remote_host  = json_config->remote_host;
        config.pear_dnsurl  = json_config->pear_dnsurl;
        config.domain_file  = json_config->domain_file;
        config.pear_dnsmasq = json_config->pear_dnsmasq;
        LOGI("[router] cp json_config end");
    }

    if (opterr) {
        usage();
        exit(EXIT_FAILURE);
    }

    if (system("killall ss-redir ss-tunnel") < 0)
    {
        LOGI("[router] killall ss-redir ss-tunnel failed!");
    }
    else
    {
        LOGI("[router] killall ss-redir ss-tunnel success!");
    }

/*
    if (config.pear_mode == 1 && (config.pear_user == NULL || config.pear_passwd == NULL))
    {
        LOGI("[ss-router] Enable 'pear_mode' need arguments: --pear_user & --pear_passwd!");
        usage();
        exit(EXIT_FAILURE);
    }
*/

    if (config.redir_port   == NULL || \
        config.remote_dns   == NULL || \
        config.tunnel_port  == NULL || \
        config.pear_dnsurl  == NULL || \
        config.pear_dnsmasq == NULL)
    {
        LOGI("[ss-router] Need arguments: redir_port || remote_dns || tunnel_port || pear_dnsurl || pear_dnsmasq !");
        usage();
        exit(EXIT_FAILURE);
    }

    if (config.pear_mode   != 1    && \
       (config.remote_num  == 0    || \
        config.remote_port == NULL || \
        config.password    == NULL ))
    {
        usage();
        exit(EXIT_FAILURE);
    }

    USE_SYSLOG(argv[0]);

    if (config.pid_flags || config.pid_path != NULL) {
        LOGI("daemonize!!");
        USE_SYSLOG(argv[0]);
        daemonize(config.pid_path);
    }


    client_redir.type  = 0;
    sprintf(client_redir.pidfile,  "%s/ss-redir.pid",  working_dir);

    client_tunnel.type = 1;
    sprintf(client_tunnel.pidfile, "%s/ss-tunnel.pid", working_dir);


    // ignore SIGPIPE
    signal(SIGPIPE, SIG_IGN);
    signal(SIGABRT, SIG_IGN);
    signal(SIGINT,  signal_cb);
    signal(SIGTERM, signal_cb);


    struct ev_loop *pear_loop = EV_DEFAULT;
    pear_ev.pear_loop = pear_loop;

    if (config.pear_mode == 1)
    {
        LOGI("enable pear auto channel mode");

        pear_fog_info_init();

        ev_timer_init(&(pear_ev.pear_timeout_channel_restart), pear_channel_restart, 1, 0);
        ev_timer_init(&(pear_ev.pear_timeout_channel_check), pear_channel_check_cb, 30, PEAR_CHANNEL_CHECK_TIME);

        if (config.pear_user != NULL && config.pear_passwd != NULL)
        {
            ev_timer_start(pear_loop, &(pear_ev.pear_timeout_channel_restart));
        }

        struct ev_timer  pear_timeout_dnslist;
        //ev_timer_init(&pear_timeout_dnslist, pear_timeout_dnslist_cb, PEAR_LIST_UPDATE_TIME, PEAR_LIST_UPDATE_TIME);
        ev_timer_init(&pear_timeout_dnslist, pear_timeout_dnslist_cb, 0, 0);
        ev_timer_start(pear_loop, &pear_timeout_dnslist);

        struct ev_timer  pear_timeout_report;

        ev_timer_init(&pear_timeout_report, pear_timeout_report_cb, 10, PEAR_ROUTER_REPORT_TIME);
/*
        ev_timer_start(pear_loop, &pear_timeout_report);
*/

    }
    else
    {
        strcpy(cloud_channel.port,   config.remote_port);
        strcpy(cloud_channel.server, config.remote_host);
        strcpy(cloud_channel.passwd, config.password);

        pear_start_client(&client_redir,  &cloud_channel);
        pear_start_client(&client_tunnel, &cloud_channel);
        
    }

    if (config.user != NULL) {
        run_as(config.user);
    }

    ev_run(pear_loop, 0);

    return 0;
}
